/*----- PROTECTED REGION ID(TwinApuId.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        TwinApuId.cpp
//
// description : C++ source for the TwinApuId class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               TwinApuId are implemented in this file.
//
// project :     twin-apu-id
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
//
// Elettra-Sincrotrone Trieste S.C.p.A. di interesse nazionale
// Strada Statale 14 - km 163,5 in AREA Science Park
// 34149 Basovizza, Trieste ITALY
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <TwinApuId.h>
#include <TwinApuIdClass.h>
#include <SupervisedID.h>
#include <algorithm>

/*----- PROTECTED REGION END -----*/	//	TwinApuId.cpp

/**
 *  TwinApuId class description:
 *    Tango device e server for Elettra Twin Apu ID (installed on TwinMic beamline) Derived from SupervisedId (cs/cls/supervisedid). The project uses git submodules
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name          |  Method name
//================================================================
//  State                 |  Inherited (no method)
//  Status                |  Inherited (no method)
//  AxisStop              |  axis_stop
//  HighEnergy            |  high_energy
//  LowEnergy             |  low_energy
//  InjectionPosition     |  injection_position
//  AxisInitLowEnergy     |  axis_init_low_energy
//  AxisInitHighEnergy    |  axis_init_high_energy
//  AxisInitEnergySwitch  |  axis_init_energy_switch
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  Enabled                |  Tango::DevBoolean	Scalar
//  Ready                  |  Tango::DevBoolean	Scalar
//  UserPositionLow        |  Tango::DevDouble	Scalar
//  UserPositionHigh       |  Tango::DevDouble	Scalar
//  HighEnergyAxisState    |  Tango::DevState	Scalar
//  LowEnergyAxisState     |  Tango::DevState	Scalar
//  EnergySwitchAxisState  |  Tango::DevState	Scalar
//  HighEnergyS            |  Tango::DevShort	Scalar
//  LowEnergyS             |  Tango::DevShort	Scalar
//  HighEnergy             |  Tango::DevBoolean	Scalar
//  LowEnergy              |  Tango::DevBoolean	Scalar
//  EnergySwitchPosition   |  Tango::DevDouble	Scalar
//  StatRF                 |  Tango::DevBoolean	Spectrum  ( max = 32)
//================================================================

namespace TwinApuId_ns
{
/*----- PROTECTED REGION ID(TwinApuId::namespace_starting) ENABLED START -----*/

//	static initializations

/*----- PROTECTED REGION END -----*/	//	TwinApuId::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : TwinApuId::TwinApuId()
 *	Description : Constructors for a Tango device
 *                implementing the classTwinApuId
 */
//--------------------------------------------------------
TwinApuId::TwinApuId(Tango::DeviceClass *cl, string &s)
 : SupervisedID(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(TwinApuId::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::constructor_1
}
//--------------------------------------------------------
TwinApuId::TwinApuId(Tango::DeviceClass *cl, const char *s)
 : SupervisedID(cl, s)
{
	/*----- PROTECTED REGION ID(TwinApuId::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::constructor_2
}
//--------------------------------------------------------
TwinApuId::TwinApuId(Tango::DeviceClass *cl, const char *s, const char *d)
 : SupervisedID(cl, s, d)
{
	/*----- PROTECTED REGION ID(TwinApuId::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : TwinApuId::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void TwinApuId::delete_device()
{
	DEBUG_STREAM << "TwinApuId::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::delete_device) ENABLED START -----*/
	//stop and join threads
	if (loop){
		loop->abortflag = true;
		//sleep(1);
		void* ptr;
		loop->join(&ptr);
		loop = 0;
	}
	if(switching_energy){
		switchenergy->abortflag = true;
		void* ptr;
		switchenergy->join(&ptr);
		switchenergy = 0;
	}
	

	//	Delete device allocated objects
	if (energy_switch_device) delete energy_switch_device;
    if (low_energy_device) delete low_energy_device;
    if (high_energy_device) delete high_energy_device;
	delete [] attr_Enabled_read;
	delete [] attr_Ready_read;
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::delete_device
	delete[] attr_UserPositionLow_read;
	delete[] attr_UserPositionHigh_read;
	delete[] attr_HighEnergyAxisState_read;
	delete[] attr_LowEnergyAxisState_read;
	delete[] attr_EnergySwitchAxisState_read;
	delete[] attr_HighEnergyS_read;
	delete[] attr_LowEnergyS_read;
	delete[] attr_HighEnergy_read;
	delete[] attr_LowEnergy_read;

	if (Tango::Util::instance()->is_svr_shutting_down()==false  &&
		Tango::Util::instance()->is_device_restarting(device_name)==false &&
		Tango::Util::instance()->is_svr_starting()==false)
	{
		//	If not shutting down call delete device for inherited object
		SupervisedID_ns::SupervisedID::delete_device();
	}
}

//--------------------------------------------------------
/**
 *	Method      : TwinApuId::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void TwinApuId::init_device()
{
	DEBUG_STREAM << "TwinApuId::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call
	high_energy_device = 0;
    low_energy_device = 0;
    energy_switch_device = 0;
    loop = 0;
    switchenergy = 0;
    switching_energy = false;

    set_state(Tango::DevState::INIT);
    set_status("init device");
    attr_EnergySwitchPosition_read = &energy_switch_position;

	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::init_device_before
	
	if (Tango::Util::instance()->is_svr_starting() == false  &&
		Tango::Util::instance()->is_device_restarting(device_name)==false)
	{
		//	If not starting up call init device for inherited object
		SupervisedID_ns::SupervisedID::init_device();
	}

	//	Get the device properties from database
	get_device_property();
	
	attr_UserPositionLow_read = new Tango::DevDouble[1];
	attr_UserPositionHigh_read = new Tango::DevDouble[1];
	attr_HighEnergyAxisState_read = new Tango::DevState[1];
	attr_LowEnergyAxisState_read = new Tango::DevState[1];
	attr_EnergySwitchAxisState_read = new Tango::DevState[1];
	attr_HighEnergyS_read = new Tango::DevShort[1];
	attr_LowEnergyS_read = new Tango::DevShort[1];
	attr_HighEnergy_read = new Tango::DevBoolean[1];
	attr_LowEnergy_read = new Tango::DevBoolean[1];
	//	No longer if mandatory property not set. 
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(TwinApuId::init_device) ENABLED START -----*/
	
	//	Initialize device
	try{
        low_energy_device = new Tango::DeviceProxy(lowEnergyAxisDeviceName);
        high_energy_device = new Tango::DeviceProxy(highEnergyAxisDeviceName);
        energy_switch_device = new Tango::DeviceProxy(energySwitchAxisDeviceName);
        DEBUG_STREAM << "created axes device proxies";
    }
    catch (Tango::DevFailed &ex){
        ERROR_STREAM << "DevFailed creating some axis device proxy";
        set_status("DevFailed creating some axis device proxy");
        set_state(Tango::DevState::FAULT);
    }
    catch (...){
        ERROR_STREAM << "unexpected exception creating some axis device proxy";
        set_status("unexpected exception creating some axis device proxy");
        set_state(Tango::DevState::FAULT);
    }
    attr_Enabled_read = new Tango::DevBoolean[1];
    attr_Enabled_read[0] = false;
    attr_Ready_read = new Tango::DevBoolean[1];
    attr_Ready_read[0] = false;
    id_flags[0]=false;
    id_flags[1]=false;
    attr_StatRF_read = id_flags;
    
    //init hostMap from hostPriviliges
	{
		vector<string>::const_iterator i;
		for (i = hostPrivileges.begin(); i != hostPrivileges.end(); i++) {
			int pos = (*i).find(':');
			if (pos <= 0) {
				ERROR_STREAM
						<< "SuperEscaID::init()  hostPrivileges syntax error "
						<< (*i) << endl;
			} else {
				string h = (*i).substr(0, pos);
				string val = (*i).substr(pos + 1);
				int valint = atoi(val.c_str());
				DEBUG_STREAM << "host: " << h << " privilege: " << valint
						<< endl;
				pair<string, int> mypair(h, valint);
				hostMap.insert(mypair);
			}
		}
	}

    Tango::MultiAttribute *attr_list = get_device_attr();	
    Tango::WAttribute	&attr_enabled  = attr_list->get_w_attr_by_name ("Enabled");	
    attr_enabled.set_write_value(attr_Enabled_write);

    //start loop thread
    loop = new readthread(this);
    loop->start();
    // TODO: init from property
    max_time_to_inj_position = maxTimeToInjectionPosition;

    set_state(Tango::DevState::ON);
    set_status("ready");
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::init_device
}

//--------------------------------------------------------
/**
 *	Method      : TwinApuId::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void TwinApuId::get_device_property()
{
	/*----- PROTECTED REGION ID(TwinApuId::get_device_property_before) ENABLED START -----*/
	
	//	Initialize property data members
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("HighEnergyAxisDeviceName"));
	dev_prop.push_back(Tango::DbDatum("LowEnergyAxisDeviceName"));
	dev_prop.push_back(Tango::DbDatum("EnergySwitchAxisDeviceName"));
	dev_prop.push_back(Tango::DbDatum("HighEnergyPosition"));
	dev_prop.push_back(Tango::DbDatum("LowEnergyPosition"));
	dev_prop.push_back(Tango::DbDatum("HighEnergyPeriod"));
	dev_prop.push_back(Tango::DbDatum("LowEnergyPeriod"));
	dev_prop.push_back(Tango::DbDatum("MaxTimeToInjectionPosition"));
	dev_prop.push_back(Tango::DbDatum("maxTimeToEnergySwitch"));
	dev_prop.push_back(Tango::DbDatum("HostPrivileges"));
	dev_prop.push_back(Tango::DbDatum("ReadThreadPollTime"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on TwinApuIdClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		TwinApuIdClass	*ds_class =
			(static_cast<TwinApuIdClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize HighEnergyAxisDeviceName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  highEnergyAxisDeviceName;
		else {
			//	Try to initialize HighEnergyAxisDeviceName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  highEnergyAxisDeviceName;
		}
		//	And try to extract HighEnergyAxisDeviceName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  highEnergyAxisDeviceName;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize LowEnergyAxisDeviceName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  lowEnergyAxisDeviceName;
		else {
			//	Try to initialize LowEnergyAxisDeviceName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  lowEnergyAxisDeviceName;
		}
		//	And try to extract LowEnergyAxisDeviceName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  lowEnergyAxisDeviceName;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize EnergySwitchAxisDeviceName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  energySwitchAxisDeviceName;
		else {
			//	Try to initialize EnergySwitchAxisDeviceName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  energySwitchAxisDeviceName;
		}
		//	And try to extract EnergySwitchAxisDeviceName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  energySwitchAxisDeviceName;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize HighEnergyPosition from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  highEnergyPosition;
		else {
			//	Try to initialize HighEnergyPosition from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  highEnergyPosition;
		}
		//	And try to extract HighEnergyPosition value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  highEnergyPosition;

		//	Try to initialize LowEnergyPosition from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  lowEnergyPosition;
		else {
			//	Try to initialize LowEnergyPosition from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  lowEnergyPosition;
		}
		//	And try to extract LowEnergyPosition value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  lowEnergyPosition;

		//	Try to initialize HighEnergyPeriod from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  highEnergyPeriod;
		else {
			//	Try to initialize HighEnergyPeriod from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  highEnergyPeriod;
		}
		//	And try to extract HighEnergyPeriod value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  highEnergyPeriod;

		//	Try to initialize LowEnergyPeriod from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  lowEnergyPeriod;
		else {
			//	Try to initialize LowEnergyPeriod from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  lowEnergyPeriod;
		}
		//	And try to extract LowEnergyPeriod value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  lowEnergyPeriod;

		//	Try to initialize MaxTimeToInjectionPosition from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  maxTimeToInjectionPosition;
		else {
			//	Try to initialize MaxTimeToInjectionPosition from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  maxTimeToInjectionPosition;
		}
		//	And try to extract MaxTimeToInjectionPosition value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  maxTimeToInjectionPosition;

		//	Try to initialize maxTimeToEnergySwitch from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  maxTimeToEnergySwitch;
		else {
			//	Try to initialize maxTimeToEnergySwitch from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  maxTimeToEnergySwitch;
		}
		//	And try to extract maxTimeToEnergySwitch value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  maxTimeToEnergySwitch;

		//	Try to initialize HostPrivileges from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  hostPrivileges;
		else {
			//	Try to initialize HostPrivileges from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  hostPrivileges;
		}
		//	And try to extract HostPrivileges value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  hostPrivileges;

		//	Try to initialize ReadThreadPollTime from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  readThreadPollTime;
		else {
			//	Try to initialize ReadThreadPollTime from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  readThreadPollTime;
		}
		//	And try to extract ReadThreadPollTime value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  readThreadPollTime;

	}

	/*----- PROTECTED REGION ID(TwinApuId::get_device_property_after) ENABLED START -----*/
	
	//	Check device property data members init
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method      : TwinApuId::check_mandatory_property()
 *	Description : For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void TwinApuId::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		append_status(tms.str());
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(TwinApuId::check_mandatory_property) ENABLED START -----*/
		cerr << tms.str() << " for " << device_name << endl;
		
		/*----- PROTECTED REGION END -----*/	//	TwinApuId::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method      : TwinApuId::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void TwinApuId::always_executed_hook()
{
	DEBUG_STREAM << "TwinApuId::always_executed_hook()  " << device_name << endl;
	if (mandatoryNotDefined)
	{
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					get_status().c_str(),
					(const char *)"TwinApuId::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(TwinApuId::always_executed_hook) ENABLED START -----*/
	
	//	code always executed before all requests
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : TwinApuId::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void TwinApuId::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "TwinApuId::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	//check state

	_state = std::max(*attr_HighEnergyAxisState_read,*attr_LowEnergyAxisState_read);
	_state = std::max(_state,*attr_EnergySwitchAxisState_read);
	set_state(_state);
	
	//chek high / low energy
	if (energy_switch_position <= lowEnergyPosition){
		*attr_LowEnergy_read = true;
		*attr_LowEnergyS_read = 1;

		*attr_HighEnergy_read = false;
		*attr_HighEnergyS_read = 0;


	}
	else if (energy_switch_position >= highEnergyPosition){
			*attr_LowEnergy_read = false;
			*attr_LowEnergyS_read = 0;

			*attr_HighEnergy_read = true;
			*attr_HighEnergyS_read = 1;
		}
	else{
		*attr_LowEnergy_read = false;
		*attr_LowEnergyS_read = 0;

		*attr_HighEnergy_read = false;
		*attr_HighEnergyS_read = 0;
	}
	//TODO: which motor states during motion? init?
	*attr_Ready_read = (*attr_LowEnergy_read ||*attr_HighEnergy_read ) && ((_state == Tango::DevState::OFF) || (_state == Tango::DevState::STANDBY));
	id_flags[0]=*attr_Ready_read;
	if (id_flags[0]) set_status("ready");
	if(switching_energy)set_status("switching energy");
	//attr_StatRF_read = id_flags;

	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : TwinApuId::write_attr_hardware()
 *	Description : Hardware writing for attributes
 */
//--------------------------------------------------------
void TwinApuId::write_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "TwinApuId::write_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::write_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute Enabled related method
 *	Description: True when
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_Enabled(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_Enabled(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_Enabled) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Enabled_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_Enabled
}
//--------------------------------------------------------
/**
 *	Write attribute Enabled related method
 *	Description: True when
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::write_Enabled(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TwinApuId::write_Enabled(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(TwinApuId::write_Enabled) ENABLED START -----*/

	int priv=caller_privileges();
	if (priv < PRIVILEGED){
		INFO_STREAM << "Enabled from "<< clientAddress <<" REJECTED"<<endl;
		TangoSys_OMemStream o;
		o << " enabled from "<< clientAddress <<" REJECTED"<<ends;
		Tango::Except::throw_exception((const char *)"Host not allowed",o.str(),(const char *)"TwinApuId::write_Enebled()");
	}


    attr_Enabled_write=w_val;
    attr_Enabled_read[0]=w_val;
    id_flags[1] = w_val;
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::write_Enabled
}
//--------------------------------------------------------
/**
 *	Read attribute Ready related method
 *	Description: true when operating condtions of the ID make it operable by the beamlines:
 *               - good flags and in range.
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_Ready(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_Ready(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_Ready) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Ready_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_Ready
}
//--------------------------------------------------------
/**
 *	Read attribute UserPositionLow related method
 *	Description: User position of low energy APU
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_UserPositionLow(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_UserPositionLow(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_UserPositionLow) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_UserPositionLow_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_UserPositionLow
}
//--------------------------------------------------------
/**
 *	Write attribute UserPositionLow related method
 *	Description: User position of low energy APU
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::write_UserPositionLow(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TwinApuId::write_UserPositionLow(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(TwinApuId::write_UserPositionLow) ENABLED START -----*/

	//check privileges
	int priv = check_privileges(AUTHORIZED, "UserPositionLow");
    //check enabled for remote useres
	if(priv == AUTHORIZED && *attr_Enabled_read == false){
		INFO_STREAM << "UserPositionLow from "<< clientAddress <<" NOT ENABLED"<<endl;
		TangoSys_OMemStream o;
		o << " UserPositionLow from "<< clientAddress <<" NOT ENABLED"<<ends;
		Tango::Except::throw_exception((const char *)"not enabled",o.str(),(const char *)"TwinApuId::UserPositionLow()");
	}

	/*check conditions*/
	if (*attr_Ready_read && *attr_LowEnergy_read) {
		try {
			Tango::DeviceAttribute da("UserPosition", w_val);
			low_energy_device->write_attribute(da);
			//force not ready while moving
			*attr_Ready_read = false;
			id_flags[0] = *attr_Ready_read;
			set_status("setting user UserPositionLow");

		} catch (Tango::DevFailed &e) {
			Tango::Except::print_exception(e);
			Tango::Except::re_throw_exception(e,
					(const char*) "Failed to set UserPositionLow on motor device",
					(const char*) "DevFailed from motor device",
					(const char*) "TwinApuId::write_UserPositionLow",
					Tango::ERR);
			set_status("failed to set UserPostionLow");

		}
	} else {
		Tango::Except::throw_exception(
				(const char*) "Failed to set UserPositionLow on motor device",
				(const char*) "Not ready or not in low energy position",
				(const char*) "TwinApuId::write_UserPositionLow", Tango::ERR);
	   set_status("failed to set UserPostionLow");
	}

	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::write_UserPositionLow
}
//--------------------------------------------------------
/**
 *	Read attribute UserPositionHigh related method
 *	Description: User position of low energy APU
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_UserPositionHigh(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_UserPositionHigh(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_UserPositionHigh) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_UserPositionHigh_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_UserPositionHigh
}
//--------------------------------------------------------
/**
 *	Write attribute UserPositionHigh related method
 *	Description: User position of low energy APU
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::write_UserPositionHigh(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TwinApuId::write_UserPositionHigh(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(TwinApuId::write_UserPositionHigh) ENABLED START -----*/
	//check privileges
	int priv = check_privileges(AUTHORIZED, "UserPositionHigh");
    //check enabled for remote useres
	if(priv == AUTHORIZED && *attr_Enabled_read == false){
		INFO_STREAM << "UserPositionHig from "<< clientAddress <<" NOT ENABLED"<<endl;
		TangoSys_OMemStream o;
		o << " UserPositionHigh from "<< clientAddress <<" NOT ENBLED"<<ends;
		Tango::Except::throw_exception((const char *)"not enabled",o.str(),(const char *)"TwinApuId::UserPositionHigh()");
	}
	/*check conditions*/
	if (*attr_Ready_read && *attr_HighEnergy_read) {
		try {
			Tango::DeviceAttribute da("UserPosition", w_val);
			high_energy_device->write_attribute(da);
			//force not ready while moving
			*attr_Ready_read = false;
			id_flags[0] = *attr_Ready_read;
			set_status("setting user UserPostionLow");
		} catch (Tango::DevFailed &e) {
			Tango::Except::print_exception(e);
			Tango::Except::re_throw_exception(e,
					(const char*) "Failed to set UserPositionHigh on motor device",
					(const char*) "DevFailed",
					(const char*) "TwinApuId::write_UserPositionHigh",
					Tango::ERR);
			set_status("failed to set UserPostionLow");

		}
	} else {
		Tango::Except::throw_exception(
				(const char*) "Failed to set UserPositionHigh on motor device",
				(const char*) "Not ready or not in high energy position",
				(const char*) "TwinApuId::write_UserPositionLow", Tango::ERR);
		set_status("failed to set UserPostionLow");
	}
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::write_UserPositionHigh
}
//--------------------------------------------------------
/**
 *	Read attribute HighEnergyAxisState related method
 *	Description: state of the HighEnergy axis device
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_HighEnergyAxisState(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_HighEnergyAxisState(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_HighEnergyAxisState) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_HighEnergyAxisState_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_HighEnergyAxisState
}
//--------------------------------------------------------
/**
 *	Read attribute LowEnergyAxisState related method
 *	Description: state of the LowEnergy axis device
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_LowEnergyAxisState(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_LowEnergyAxisState(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_LowEnergyAxisState) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_LowEnergyAxisState_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_LowEnergyAxisState
}
//--------------------------------------------------------
/**
 *	Read attribute EnergySwitchAxisState related method
 *	Description: state of the EnergySwitch axis device
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_EnergySwitchAxisState(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_EnergySwitchAxisState(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_EnergySwitchAxisState) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_EnergySwitchAxisState_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_EnergySwitchAxisState
}
//--------------------------------------------------------
/**
 *	Read attribute HighEnergyS related method
 *	Description: 1 if the EnergySwitch Axes is int the HighEnergy Position
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_HighEnergyS(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_HighEnergyS(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_HighEnergyS) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_HighEnergyS_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_HighEnergyS
}
//--------------------------------------------------------
/**
 *	Read attribute LowEnergyS related method
 *	Description: 1 if the EnergySwitch Axes is int the LowEnergy Position
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_LowEnergyS(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_LowEnergyS(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_LowEnergyS) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_LowEnergyS_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_LowEnergyS
}
//--------------------------------------------------------
/**
 *	Read attribute HighEnergy related method
 *	Description: tue if the EnergySwitch Axes is int the HighEnergy Position
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_HighEnergy(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_HighEnergy(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_HighEnergy) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_HighEnergy_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_HighEnergy
}
//--------------------------------------------------------
/**
 *	Read attribute LowEnergy related method
 *	Description: 1 if the EnergySwitch Axes is int the LowEnergy Position
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_LowEnergy(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_LowEnergy(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_LowEnergy) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_LowEnergy_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_LowEnergy
}
//--------------------------------------------------------
/**
 *	Read attribute EnergySwitchPosition related method
 *	Description: UserPosition of energy switch motor
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TwinApuId::read_EnergySwitchPosition(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_EnergySwitchPosition(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_EnergySwitchPosition) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_EnergySwitchPosition_read);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_EnergySwitchPosition
}
//--------------------------------------------------------
/**
 *	Read attribute StatRF related method
 *	Description: bit mask for compatibility with old style software
 *               Holds the ready and enabled bits together
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Spectrum max = 32
 */
//--------------------------------------------------------
void TwinApuId::read_StatRF(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TwinApuId::read_StatRF(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TwinApuId::read_StatRF) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_StatRF_read, 2);
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::read_StatRF
}

//--------------------------------------------------------
/**
 *	Method      : TwinApuId::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void TwinApuId::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(TwinApuId::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command AxisStop related method
 *	Description: send AxisStop to all axes (motors)
 *
 */
//--------------------------------------------------------
void TwinApuId::axis_stop()
{
	DEBUG_STREAM << "TwinApuId::AxisStop()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::axis_stop) ENABLED START -----*/
	
	//	Add your own code

	if(switching_energy){
		switchenergy->abortflag = true;
		void* ptr;
		switchenergy->join(&ptr);
		switchenergy = 0;
	}
	try {
		high_energy_device->command_inout("AxisStop");
		low_energy_device->command_inout("AxisStop");
		energy_switch_device->command_inout("AxisStop");
		*attr_Ready_read = false;
		id_flags[0]=*attr_Ready_read;
	} catch (Tango::DevFailed &e) {
		Tango::Except::print_exception(e);
		Tango::Except::re_throw_exception(e,
				(const char*) "Failed to Stop motor device",
				(const char*) "DevFailed", (const char*) "TwinApuId::axis_stop",
				Tango::ERR);
	}
	
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::axis_stop
}
//--------------------------------------------------------
/**
 *	Command HighEnergy related method
 *	Description: activate the high energy APU
 *
 */
//--------------------------------------------------------
void TwinApuId::high_energy()
{
	DEBUG_STREAM << "TwinApuId::HighEnergy()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::high_energy) ENABLED START -----*/
	//	Add your own code
	check_privileges(PRIVILEGED,"high_energy");
	if(*attr_HighEnergy_read) return;
	switchenergy = new switchenergythread(this,true);
	switchenergy->start();
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::high_energy
}
//--------------------------------------------------------
/**
 *	Command LowEnergy related method
 *	Description: activate the low energy APU
 *
 */
//--------------------------------------------------------
void TwinApuId::low_energy()
{
	DEBUG_STREAM << "TwinApuId::LowEnergy()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::low_energy) ENABLED START -----*/

	//	Add your own code
	check_privileges(PRIVILEGED,"low_energy");
	if(*attr_LowEnergy_read) return;
	switchenergy = new switchenergythread(this,false);
	switchenergy->start();
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::low_energy
}
//--------------------------------------------------------
/**
 *	Command InjectionPosition related method
 *	Description: move both APUs to Injection position
 *
 */
//--------------------------------------------------------
void TwinApuId::injection_position()
{
	DEBUG_STREAM << "TwinApuId::InjectionPosition()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::injection_position) ENABLED START -----*/
	
    //	Add your own code
	check_privileges(PRIVILEGED, "InjectionPosition");
    try {
        double inj_pos_high = highEnergyPeriod/2.0;
        double inj_pos_low = lowEnergyPeriod/2.0;
        Tango::DeviceAttribute da_high("UserPosition", inj_pos_high);
        Tango::DeviceAttribute da_low("UserPosition", inj_pos_low);
        high_energy_device->write_attribute(da_high);
        low_energy_device->write_attribute(da_low);
        //force not ready while moving
        *attr_Ready_read = false;
        id_flags[0] = *attr_Ready_read;
        //attr_StatRF_read = id_flags;
    } catch (Tango::DevFailed &e) {
    	Tango::Except::print_exception(e);
        Tango::Except::re_throw_exception(e,
                                          (const char*) "Failed to set injection position on motor device",
                                          (const char*) "DevFailed",
                                          (const char*) "TwinApuId::injection_position",
                                          Tango::ERR);
        
    }
    /*----- PROTECTED REGION END -----*/	//	TwinApuId::injection_position
}
//--------------------------------------------------------
/**
 *	Command AxisInitLowEnergy related method
 *	Description: send AxisInit to LowEnergy motor
 *
 */
//--------------------------------------------------------
void TwinApuId::axis_init_low_energy()
{
	DEBUG_STREAM << "TwinApuId::AxisInitLowEnergy()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::axis_init_low_energy) ENABLED START -----*/
	
	//	Add your own code
	check_privileges(PRIVILEGED, "AxisIniLowEnergy");
	try {
		low_energy_device->command_inout("AxisInit");
		*attr_Ready_read = false;
		id_flags[0]=*attr_Ready_read;
		set_status("AxisInit");
	} catch (Tango::DevFailed &e) {
		Tango::Except::print_exception(e);
		Tango::Except::re_throw_exception(e,
				(const char*) "Failed to AxisInitLowEnergy motor device",
				(const char*) "DevFailed", (const char*) "TwinApuId::axis_init_low_energy",
				Tango::ERR);
		set_status("AxisInitLowEnergy failed");
	}
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::axis_init_low_energy
}
//--------------------------------------------------------
/**
 *	Command AxisInitHighEnergy related method
 *	Description: send AxisInit to HighEnergy motor
 *
 */
//--------------------------------------------------------
void TwinApuId::axis_init_high_energy()
{
	DEBUG_STREAM << "TwinApuId::AxisInitHighEnergy()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::axis_init_high_energy) ENABLED START -----*/
	
	//	Add your own code
	check_privileges(PRIVILEGED, "AxisInitHighEnergy");
	try {
		high_energy_device->command_inout("AxisInit");
		*attr_Ready_read = false;
		id_flags[0]=*attr_Ready_read;
		set_status("AxisInit");
	} catch (Tango::DevFailed &e) {
		Tango::Except::print_exception(e);
		Tango::Except::re_throw_exception(e,
				(const char*) "Failed to AxisInitHighEnergy motor device",
				(const char*) "DevFailed", (const char*) "TwinApuId::axis_init_high_energy",
				Tango::ERR);
		set_status("AxisInit HighEnergy failed");
	}
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::axis_init_high_energy
}
//--------------------------------------------------------
/**
 *	Command AxisInitEnergySwitch related method
 *	Description: send AxisInit to Energy Switch motor
 *
 */
//--------------------------------------------------------
void TwinApuId::axis_init_energy_switch()
{
	DEBUG_STREAM << "TwinApuId::AxisInitEnergySwitch()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TwinApuId::axis_init_energy_switch) ENABLED START -----*/

	//	Add your own code
	check_privileges(PRIVILEGED, "AxisInitEnergySwitch");
	try {
		energy_switch_device->command_inout("AxisInit");
		*attr_Ready_read = false;
		id_flags[0]=*attr_Ready_read;
		set_status("AxisInit");
	} catch (Tango::DevFailed &e) {
		Tango::Except::print_exception(e);
		Tango::Except::re_throw_exception(e,
				(const char*) "Failed to AxisInit motor device",
				(const char*) "DevFailed", (const char*) "TwinApuId::axis_init_energy_switch",
				Tango::ERR);
		set_status("AxisInitEnergySwitch failed");
	}
	/*----- PROTECTED REGION END -----*/	//	TwinApuId::axis_init_energy_switch
}
//--------------------------------------------------------
/**
 *	Method      : TwinApuId::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void TwinApuId::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(TwinApuId::add_dynamic_commands) ENABLED START -----*/

	//	Add your own code to create and add dynamic commands if any

	/*----- PROTECTED REGION END -----*/	//	TwinApuId::add_dynamic_commands
}

/*----- PROTECTED REGION ID(TwinApuId::namespace_ending) ENABLED START -----*/

//	Additional Methods
//-------------------------------------------------------------------------------
    /**
	 * assign privileges elevel based on host name from which the call is generated
	 * PRIVILEGED : can change the enable bit and set gap/tapering
	 * AUTHORIZED : can set gap/tapering when enabled
	 * NORMAL     : can only read;
	 */
int TwinApuId::caller_privileges(void)
{
	//if no privileges are configured, run unimpeded
	if(hostMap.size() == 0)return PRIVILEGED;

	Tango::DevVarStringArray blackbox = black_box(1)[0];
	string last_msg(blackbox[0].in());
	string::size_type loc = last_msg.find( " requested from ", 0 );
	if( loc != string::npos )
	{
		clientAddress = last_msg.substr(loc + strlen(" requested from "));
		loc=clientAddress.find(" ",0);
		clientAddress =clientAddress.substr(0,loc);
		INFO_STREAM<<"checking call from "<<clientAddress<<endl;
		//         std::set<std::string> client_ips;
		//         get_ips_from_host(clientAddress, client_ips);
		//
		//         for (std::set<std::string>::const_iterator it = client_ips.begin(); it != client_ips.end(); ++it )
		//         {
		//             if (std::find(allowed_ips.begin(), allowed_ips.end(), *it) != allowed_ips.end())
		//                 return;  // found the IP in the allowed list of IPs
		//         }
		map<string,int>::iterator it;
		it=hostMap.find(clientAddress);
		if (it == hostMap.end()) {
			INFO_STREAM << clientAddress<< " host not foud" << endl;
			return NORMAL;
		}
		else {
			INFO_STREAM << "host: "<< clientAddress<< " privilges="<< it->second << endl;
			return it->second;
		}

	}
	else
	{
		WARN_STREAM << "Cannot extract client IP or hostname from tango blackbox, denying access.";
		clientAddress="";
		return NORMAL; //do not grant special privileges
	}
	return NORMAL; //do not grant special privilege - but we should NOT arrive at this line
}
//--------------------------------------------------------------------
/**
	 * check privileges level based on host name from which the call is generated
	 * formats warning/  error messages and throw exception in case of vilotaions
	 * privilege levels:
	 * PRIVILEGED : can change the enable bit and set gap/tapering
	 * AUTHORIZED : can set gap/tapering when enabled
	 * NORMAL     : can only read;
	 */

int TwinApuId::check_privileges(int level, string msg)
{
	int priv=caller_privileges();
	if (priv < level){
		INFO_STREAM << msg << " from "<< clientAddress <<" REJECTED"<<endl;
		TangoSys_OMemStream o;
		o << msg << " from "<< clientAddress <<" REJECTED"<<ends;
		Tango::Except::throw_exception((const char *)"Host not allowed",o.str(),msg);
	}
	return priv;
}

/*----- PROTECTED REGION END -----*/	//	TwinApuId::namespace_ending
} //	namespace
